/** Z80: portable Z80 emulator *******************************/
/**                                                         **/
/**                           Z80.h                         **/
/**                                                         **/
/** This file contains declarations relevant to emulation   **/
/** of Z80 CPU.                                             **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1994,1995,1996,1997       **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/   
/**     changes to this file.                               **/
/*************************************************************/
#ifndef Z80_H
#define Z80_H

#include "solace_types.h"

                               /* Compilation options:       */
/* #define DEBUG */            /* Compile debugging version  */
#define LSB_FIRST        /* Compile for low-endian CPU */
/* #define MSB_FIRST */        /* Compile for hi-endian CPU  */

                               /* LoopZ80() may return:      */
#define INT_IRQ     0x0038     /* Standard RST 38h interrupt */
#define INT_NMI     0x0066     /* Non-maskable interrupt     */
#define INT_NONE    0xFFFF     /* No interrupt required      */
#define INT_QUIT    0xFFFE     /* Exit the emulation         */

                               /* Bits in Z80 F register:    */
#define S_FLAG      0x80       /* 1: Result negative         */
#define Z_FLAG      0x40       /* 1: Result is zero          */
#define H_FLAG      0x10       /* 1: Halfcarry/Halfborrow    */
#define P_FLAG      0x04       /* 1: Result is even          */
#define V_FLAG      0x04       /* 1: Overflow occured        */
#define N_FLAG      0x02       /* 1: Subtraction occured     */
#define C_FLAG      0x01       /* 1: Carry/Borrow occured    */


/** Structured Datatypes *************************************/
/** NOTICE: #define LSB_FIRST for machines where least      **/
/**         signifcant byte goes first.                     **/
/*************************************************************/
typedef union
{
#ifdef LSB_FIRST
  struct { byte l,h; } B;
#else
  struct { byte h,l; } B;
#endif
  word W;
} pair;


typedef struct
{
  pair AF,BC,DE,HL,PC,SP;             /* Main registers      */
  byte IFF;                           /* Interrupt registers */

  int ICount;         /* cycle count, backup cycle count     */
  int IBackup;        /* Private, don't touch                */
  word IRequest;      /* Set to address of pending IRQ       */
  void *User;         /* Arbitrary user data (ID,RAM*,etc.)  */

  /* jtb solace enhancements follow */
  pair startPC;       /* if debugging, PC of 1st byte of op  */
  int  brkpt;         /* set if breakpoint is detected       */
  int  brknum;        /* serial number of breakpoint         */
  brkpt_t reason;     /* reason why breakpoint occurred      */

} Z80;

/** ResetZ80() ***********************************************/
/** This function can be used to reset the registers before **/
/** starting execution with RunZ80(). It sets registers to  **/
/** their initial values.                                   **/
/*************************************************************/
void ResetZ80(Z80 *R);

/** ExecZ80() ************************************************/
/** This function will execute a single Z80 opcode. It will **/
/** then return next PC, and current register values in R.  **/
/*************************************************************/
//word ExecZ80(Z80 *R);
word Exec8080(Z80 *R);

/** RunZ80() *************************************************/
/** This function will run Z80 code until an LoopZ80() call **/
/** returns INT_QUIT. It will return the PC at which        **/
/** emulation stopped, and current register values in R.    **/
/*************************************************************/
// JTB:
/** This function will run Z80 code until the R->ICount     **/
/** value is exhausted.                                     **/
//word RunZ80(Z80 *R);
//word Run8080(Z80 *R);

/** RdZ80()/WrZ80() ******************************************/
/** These functions are called when access to RAM occurs.   **/
/** They allow to control memory access.                    **/
/************************************ TO BE WRITTEN BY USER **/
//void WrZ80(register word Addr,register byte Value);
//byte RdZ80(register word Addr);

/** InZ80()/OutZ80() *****************************************/
/** Z80 emulation calls these functions to read/write from  **/
/** I/O ports. There can be 65536 I/O ports, but only first **/
/** 256 are usually used                                    **/
/************************************ TO BE WRITTEN BY USER **/
//void OutZ80(word Port, byte Value);
//byte InZ80(word Port);

/** PatchZ80() ***********************************************/
/** Z80 emulation calls this function when it encounters a  **/
/** special patch command (ED FE) provided for user needs.  **/
/** For example, it can be called to emulate BIOS calls,    **/
/** such as disk and tape access. Replace it with an empty  **/
/** macro for no patching.                                  **/
/************************************ TO BE WRITTEN BY USER **/
//void PatchZ80(Z80 *R);

/** DebugZ80() ***********************************************/
/** This function should exist if DEBUG is #defined. When   **/
/** Trace!=0, it is called after each command executed by   **/
/** the CPU, and given the Z80 registers. Emulation exits   **/
/** if DebugZ80() returns 0.                                **/
/*************************************************************/
//byte DebugZ80(Z80 *R);

/** LoopZ80() ************************************************/
/** Z80 emulation calls this function periodically to check **/
/** if the system hardware requires any interrupts. This    **/
/** function must return an address of the interrupt vector **/
/** (0x0038, 0x0066, etc.) or INT_NONE for no interrupt.    **/
/** Return INT_QUIT to exit the emulation loop.             **/
/************************************ TO BE WRITTEN BY USER **/
//word LoopZ80(Z80 *R);

/** System-Dependent Stuff ***********************************/
/** This is system-dependent code put here to speed things  **/
/** up. It has to stay inlined to be fast.                  **/
/*************************************************************/

// DiZ80 is used by the disassembler to read a byte.
#define DiZ80(Addr)       (*simstate.diz80_func)(Addr)
// OpZ80 is used by the simulator to fetch the first opcode byte.
#define OpZ80(Addr)       (*simstate.opz80_func)(Addr)
// RdZ80 is used by the simulator to fetch 2nd-4th opcode and data bytes
#define RdZ80(Addr)       (*simstate.rdz80_func)(Addr)
// WrZ80 is used by the simulator to write bytes
#define WrZ80(Addr,Value) (*simstate.wrz80_func)((Addr),(Value))
// InZ80 is used by the simulator for the IN opcode
#define InZ80(Addr)       (*simstate.inz80_func)(Addr)
// OutZ80 is used by the simulator for the OUT opcode
#define OutZ80(Addr,Value) (*simstate.outz80_func)((Addr),(Value))

#endif /* Z80_H */
