// Solace -- Sol Anachronistic Computer Emulation
// A Win32 emulator for the Sol-20 computer.
//
// Copyright (c) Jim Battle, 2000, 2001

// ===============================================================
// The following section contains code for the disassembler window.
// The window has the following behaviors:
//     1) can be "disabled" (while CPU is running)
//     2) supports normal scrollbar behavior
//     3) supports uparrow/downarrow/pageup/pagedown/home keys
//     4) the left column indicates lines that have breakpoints
//     5) the current line is in bold
//     6) double clicking on a line causes the CPU to begin
//        running and sets a breakpoint on the line clicked on
//     7) right clicking brings up a breakpoint dialog

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <windows.h>
#include <windowsx.h>	// message cracker macros

#include "wingui.h"
#include "windbg_sub.h"
#include "resource.h"
#include "solace_intf.h"
#include "z80.h"


static struct {

    BOOL active;	// FALSE=disabled
    WORD toppc;		// PC of instruction at top of window
    int  win_x;		// window width, in pixels
    int  chars_x;	// # chars wide (rounded up)
    int  chars_y;	// # chars high (rounded up)
    int  *ops;		// address of each op of each line
    int  xoff;		// horizontal scrolling offset

    // this is ugly, but here goes.  when the right mouse button
    // goes down to create the popup menu, the handler figures out
    // which PC the mouse is clicking on, and tucks it away here.
    // later when the menu command is received by that WM_COMMAND
    // handler, it looks at it here.
    word rightbut_addr;
    word rightbut_brknum;

} dasmstate;


// given the address of the first op in the window,
// calculate the instruction address for each line
static void
RegenerateOps(word pc)
{
    int i;

    for(i=0; i<dasmstate.chars_y; i++) {
	int oplen = OpLen(pc);
	dasmstate.ops[i] = pc;
	pc += oplen;
    }
}


static BOOL
OnCreateDasm(HWND hwnd, LPCREATESTRUCT lpCreateStruct)
{
    int pc = CPU_RegRead(CPU_REG_PC);

    // set range for scrollbars
    // vertical runs from 0000 to FFFF
    SetScrollRange(hwnd, SB_VERT, 0x0000, 0xFFFF, FALSE);
    SetScrollPos(hwnd, SB_VERT, pc, TRUE);

    return TRUE;
}


// handle window resize events
static void
OnSizeDasm(HWND hwnd, UINT state, int cx, int cy)
{
    dasmstate.win_x = cx;

    dasmstate.chars_x = (cx + dbgstate.fix_x - 1) / dbgstate.fix_x;
    dasmstate.chars_y = (cy + dbgstate.fix_y - 1) / dbgstate.fix_y;
    if (dasmstate.chars_y < 1)
	dasmstate.chars_y = 1;

    if (dasmstate.chars_x < 50) {
	SetScrollRange(hwnd, SB_HORZ, 0, 50-dasmstate.chars_x, FALSE);
	ShowScrollBar(hwnd, SB_HORZ, TRUE);
    } else {
	ShowScrollBar(hwnd, SB_HORZ, FALSE);
	dasmstate.xoff = 0;	// in case we had one coming in
    }

    SetScrollPos(hwnd, SB_HORZ, 0, TRUE);

    // store the starting PC for each instruction in the window
    if (dasmstate.ops != NULL)
	free(dasmstate.ops);
    dasmstate.ops = malloc(dasmstate.chars_y * sizeof(int));
    ASSERT(dasmstate.ops != NULL);
    RegenerateOps(dasmstate.toppc);
}


static void
OnPaintDasm(HWND hwnd)
{
    PAINTSTRUCT	ps;
    HDC hdc;
    int i;

    hdc = BeginPaint(hwnd, &ps);

    if (!dasmstate.active) {

	// gray out the surface
	GrayOutWindow(hwnd, hdc);

    } else {

	int saved = SaveDC(hdc);

	COLORREF white  = RGB(0xFF, 0xFF, 0xFF);
	COLORREF yellow = RGB(0xFF, 0xFF, 0x00);
	COLORREF red    = RGB(0xFF, 0x00, 0x00);

	HBRUSH whitebrush  = CreateSolidBrush(      white);
	HPEN   whitepen    = CreatePen(PS_SOLID, 1, white);
	HBRUSH yellowbrush = CreateSolidBrush(      yellow);
	HPEN   yellowpen   = CreatePen(PS_SOLID, 1, yellow);
	HBRUSH redbrush    = CreateSolidBrush(      red);
	HPEN   redpen      = CreatePen(PS_SOLID, 1, red);

	SelectFont(hdc, dbgstate.fixed_font);

	for(i=0; i<dasmstate.chars_y; i++) {

	    char dasmbuf[80], dasmbuf2[80];
	    word dasmpc = dasmstate.ops[i];
	    BOOL curpc = (dasmpc == CPU_RegRead(CPU_REG_PC));
	    int bp     = check_brk(dasmpc, BRK_OP);
#if 0
	    // if we are disassembling the current pc,
	    // this will not only include the mnemonic,
	    // but the address operated on by an indirect
	    // instruction, and the data item pointed at
	    // in the case of a read.  however, this screws
	    // things up in the case that the memory location
	    // has side effects on read, since the location gets
	    // read once at disassembly and once at actual execution.
	    // FIXME: perhaps change DAsm to know which locations
	    // are safe to read.
	    int oplen = DAsm(dasmbuf2, dasmpc, curpc);
#else
	    int oplen = DAsm(dasmbuf2, dasmpc, 0);
#endif
	    int len, j;

	    sprintf(dasmbuf, "  %04X ", dasmpc);  // left 2 spaces open

	    for(j=0; j<3; j++) {
		if (j < oplen)
		    sprintf(&dasmbuf[7+3*j], "%02X ", Sys_ReadMappedByte((word)(dasmpc+j)));
		else
		    sprintf(&dasmbuf[7+3*j], "   ");
	    }

	    len = 16 + sprintf(&dasmbuf[16], " %s", dasmbuf2);

	    len -= dasmstate.xoff;
	    if (len < 0)
		len = 0;

	    if (curpc) {
		SetBkColor(hdc, yellow);
		SelectObject(hdc, yellowbrush);
		SelectObject(hdc, yellowpen);
	    } else {
		SetBkColor(hdc, white);
		SelectObject(hdc, whitebrush);
		SelectObject(hdc, whitepen);
	    }

	    // erase from the end of the text to the RHS of the window
	    Rectangle(hdc,
		      len*dbgstate.fix_x, i*dbgstate.fix_y,	// l, t
		      dasmstate.win_x, (i+1)*dbgstate.fix_y);	// r, b

	    if (len > 0)
		TextOut(hdc, 0, i*dbgstate.fix_y,
			     &dasmbuf[dasmstate.xoff], len);

	    // if dasmpc == breakpoint, draw stop sign in left column
	    if (bp) {

		// calculate circle coordinates
		int dia = min(dbgstate.fix_x, dbgstate.fix_y);
		int ctr_2y = ((i+i+1)*dbgstate.fix_y);
		int ctr_2x = 2*dbgstate.fix_x;

		SelectObject(hdc, redpen);

		if (bp > 0)
		    SelectObject(hdc, redbrush);

		Ellipse(hdc,
			(ctr_2x - dia)/2, (ctr_2y - dia)/2,	// l, t
			(ctr_2x + dia)/2, (ctr_2y + dia)/2);	// r, b
	    } // if (bp)

	} // for i

	RestoreDC(hdc, saved);

	DeleteObject(whitebrush);
	DeleteObject(whitepen);
	DeleteObject(yellowbrush);
	DeleteObject(yellowpen);
	DeleteObject(redbrush);
	DeleteObject(redpen);
    }

    EndPaint(hwnd, &ps);
}


static void
OnKeyDownDasm(HWND hwnd, UINT vk, BOOL down, int repeat, UINT flags)
{
    word pc;

    switch (vk) {

	case VK_UP:    DasmWindowAction(WA_LINEUP, 0);          return;
	case VK_DOWN:  DasmWindowAction(WA_LINEDOWN, 0);        return;
	case VK_PRIOR: DasmWindowAction(WA_PAGEUP, 0);          return;
	case VK_NEXT:  DasmWindowAction(WA_PAGEDOWN, 0);        return;
	case VK_HOME:
	    pc = (word)CPU_RegRead(CPU_REG_PC);
	    DasmWindowAction(WA_GOTO, pc);
	    return;

	case VK_F5:
	case VK_F6:
	case VK_F7:
	    handle_fkey(vk);
	    return;

	default: break;
    }

    // use default message handler
    FORWARD_WM_KEYDOWN(hwnd, vk, repeat, flags, DefWindowProc);
}


static void
OnDestroyDasm(HWND hwnd)
{
    if (dasmstate.ops != NULL) {
	free(dasmstate.ops);
	dasmstate.ops = NULL;
    }

    dbgstate.hDasmWnd = NULL;
}


static void
OnVScrollDasm(HWND hwnd, HWND hctl, UINT code, int pos)
{
    word pc;

    SetFocus(hwnd);	// make keyboard input go to this subwin

    switch (code) {

	case SB_LINEUP:   DasmWindowAction(WA_LINEUP,   0); break;
	case SB_LINEDOWN: DasmWindowAction(WA_LINEDOWN, 0); break;
	case SB_PAGEUP:   DasmWindowAction(WA_PAGEUP,   0); break;
	case SB_PAGEDOWN: DasmWindowAction(WA_PAGEDOWN, 0); break;

	case SB_THUMBTRACK:
	    {
		// show current value in real time, centered
		HDC hdc = GetDC(hwnd);
		COLORREF bg       = RGB(0x00, 0xFF, 0xFF);
		COLORREF oldcolor = SetBkColor(hdc, bg);
		char buf[10];
		int len = sprintf(buf, " %04X  ", (pos & 0xFFFF));

		TextOut(hdc, dbgstate.fix_x * (dasmstate.chars_x-4)/2,
			     dbgstate.fix_y * (dasmstate.chars_y)  /2,
			     &buf[dasmstate.xoff], len);

		SetBkColor(hdc, oldcolor);
		ReleaseDC(hwnd, hdc);
	    }
	    return;

	case SB_THUMBPOSITION:
	    // the position reported might be in the middle of a real op.
	    // Perhaps the best thing to do is to back up N instructions
	    // and start at the nearest real op.
	    pos &= 0xFFFF;	// although passed as int, really signed short
	    pc = (pos < 3*10) ? 0 : (pos - 3*10);	// 10 max length ops
	    while (pc < pos) {
		int len = OpLen(pc);
		if (pc+len > pos)
		    break;
		pc += len;
	    }
	    SetScrollPos(hwnd, SB_VERT, (int)pc, TRUE);
	    DasmWindowAction(WA_GOTO, pc);
	    return;
	default: return;
    }

    // get current top line and set scrollbar position
    pc = dasmstate.ops[0];
    SetScrollPos(hwnd, SB_VERT, (int)pc, TRUE);
}


static void
OnHScrollDasm(HWND hwnd, HWND hctl, UINT code, int pos)
{
    SetFocus(hwnd);	// make keyboard input go to this subwin

    switch (code) {

	case SB_LINEUP:   dasmstate.xoff--; break;
	case SB_LINEDOWN: dasmstate.xoff++; break;
	case SB_PAGEUP:   dasmstate.xoff -= dasmstate.chars_x/2; break;
	case SB_PAGEDOWN: dasmstate.xoff += dasmstate.chars_x/2; break;
#if 1
	case SB_THUMBTRACK:
#else
	case SB_THUMBPOSITION:
#endif
	    dasmstate.xoff = pos;
	    break;
	default: return;
    }

    if (dasmstate.xoff > 80-dasmstate.chars_x)
	dasmstate.xoff = 80-dasmstate.chars_x;

    if (dasmstate.xoff < 0)
	dasmstate.xoff = 0;

    SetScrollPos(hwnd, SB_HORZ, (int)dasmstate.xoff, TRUE);
    InvalidateRect(hwnd, NULL, TRUE);
}


// if we control click on a line, we cycle between three states:
//     no breakpoint, enabled breakpoint, disabled breakpoint
static void
OnLButtonUpDasm(HWND hWnd, int xClient, int yClient, UINT keyflags)
{
    int row = yClient / dbgstate.fix_y;
    int brkstate;
    word addr;
    char cmdstr[40];

    if (~keyflags & MK_CONTROL) {
	// control key isn't pressed; just set focus
	SetFocus(hWnd);
	return;
    }

    // figure out which address corresponds to that row.
    // it is a bit dirty to directly access the window state
    // instead of through an abstraction, but what the hell...
    addr = (word)dasmstate.ops[row];

    brkstate = check_brk(addr, BRK_OP);
    if (brkstate == 0) {
	// no breakpoint -- so create one
	sprintf(cmdstr, "bpc %04x", addr);
    } else if (brkstate < 0) {
	// disabled breakpoint exists; remove it
	sprintf(cmdstr, "bkill #%d", -brkstate);
    } else /* (brkstate > 0) */ {
	// active breakpoint exists; disable it
	sprintf(cmdstr, "bdisable #%d", brkstate);
    }

    UI_DbgWinLog(cmdstr, TRUE);	// echo it
    dbg_interp(cmdstr);		// interpret it
}


// right clicking in the Dasm window creates a pop-up menu
static void
OnRButtonDownDasm(HWND hWnd, BOOL dblclk, int xClient, int yClient, UINT keyflags)
{
    int row = yClient / dbgstate.fix_y;
    int brkstate;
    word addr;
    HMENU hmenu;
    POINT mousepos;
    UINT greyed;
    char str[40];

    SetFocus(hWnd);

    hmenu = CreatePopupMenu();
    if (!hmenu)
	return;	// oops

    // figure out which address corresponds to that row.
    // it is a bit dirty to directly access the window state
    // instead of through an abstraction, but what the hell...
    addr = (word)dasmstate.ops[row];
    dasmstate.rightbut_addr   = addr;

    brkstate = check_brk(addr, BRK_OP);
    dasmstate.rightbut_brknum = (brkstate < 0) ? -brkstate : brkstate;

    greyed = (brkstate == 0) ? MF_ENABLED : MF_GRAYED;
    sprintf(str, "&Add Breakpoint at pc=%04X", addr);
    AppendMenu(hmenu, MF_STRING | greyed, ID_ADDBREAKPOINT,     str);

    greyed = (brkstate >  0) ? MF_ENABLED : MF_GRAYED;
    sprintf(str, "&Disable Breakpoint #%d (%04X)", dasmstate.rightbut_brknum, dasmstate.rightbut_addr);
    AppendMenu(hmenu, MF_STRING | greyed, ID_DISABLEBREAKPOINT, str);

    greyed = (brkstate <  0) ? MF_ENABLED : MF_GRAYED;
    sprintf(str, "&Enable Breakpoint #%d (%04X)", dasmstate.rightbut_brknum, dasmstate.rightbut_addr);
    AppendMenu(hmenu, MF_STRING | greyed, ID_ENABLEBREAKPOINT,  str);

    greyed = (brkstate != 0) ? MF_ENABLED : MF_GRAYED;
    sprintf(str, "&Kill Breakpoint #%d (%04X)", dasmstate.rightbut_brknum, dasmstate.rightbut_addr);
    AppendMenu(hmenu, MF_STRING | greyed, ID_KILLBREAKPOINT,    str);

    AppendMenu(hmenu, MF_STRING | MF_SEPARATOR, 0,     NULL);

    sprintf(str, "&Run to pc=%04X", addr);
    AppendMenu(hmenu, MF_STRING, ID_RUNBREAKPOINT,     str);

    AppendMenu(hmenu, MF_STRING | MF_SEPARATOR, 0,     NULL);

    AppendMenu(hmenu, MF_STRING, ID_HOMEPC,            "&Home current PC");

    // convert mouse coords to screen coords for use by TrackPopupMenu
    mousepos.x = xClient;
    mousepos.y = yClient;
    ClientToScreen(hWnd, &mousepos);

    TrackPopupMenu(hmenu, TPM_LEFTALIGN | TPM_RIGHTBUTTON,
		    mousepos.x, mousepos.y, 0, hWnd, NULL);

    DestroyMenu(hmenu);
}


// double clicking on a line sets a temporary breakpoint on the
// specified instruction.
static void
OnLButtonDblClkDasm(HWND hWnd, BOOL dblclk, int xClient, int yClient, UINT keyflags)
{
    int row = yClient / dbgstate.fix_y;
    word addr;
    char cmdstr[40];

    if (keyflags & MK_CONTROL)
	return;	// control key is pressed

    addr = (word)dasmstate.ops[row];

    sprintf(cmdstr, "to %04x", addr);
    UI_DbgWinLog(cmdstr, TRUE);	// echo it
    dbg_interp(cmdstr);		// interpret it
}


static void
OnCommandDasm(HWND hwnd, int id, HWND hctl, UINT codenotify)
{
    char cmdstr[40];

    switch (id) {
	case ID_ADDBREAKPOINT:
	    sprintf(cmdstr, "bpc %04x", dasmstate.rightbut_addr);
	    break;
	case ID_DISABLEBREAKPOINT:
	    sprintf(cmdstr, "bdisable #%d", dasmstate.rightbut_brknum);
	    break;
	case ID_ENABLEBREAKPOINT:
	    sprintf(cmdstr, "benable #%d", dasmstate.rightbut_brknum);
	    break;
	case ID_KILLBREAKPOINT:
	    sprintf(cmdstr, "bkill #%d", dasmstate.rightbut_brknum);
	    break;
	case ID_RUNBREAKPOINT:
	    sprintf(cmdstr, "to %04x", dasmstate.rightbut_addr);
	    break;
	case ID_HOMEPC:
	    DasmWindowAction(WA_GOTO, (word)CPU_RegRead(CPU_REG_PC));
	    return;
	default:
	    return;
    }
    UI_DbgWinLog(cmdstr, TRUE);	// echo it
    dbg_interp(cmdstr);		// interpret it

    // we aren't handling it, pass it through to default handler
    FORWARD_WM_COMMAND(hwnd, id, hctl, codenotify, DefWindowProc);
}


// debugger disassembler window handler
static LRESULT CALLBACK
WndProcDasm(HWND hWnd, UINT iMsg, WPARAM wParam, LPARAM lParam)
{
    switch (iMsg) {

	HANDLE_MSG(hWnd, WM_CREATE,        OnCreateDasm);
	HANDLE_MSG(hWnd, WM_SIZE,          OnSizeDasm);
	HANDLE_MSG(hWnd, WM_PAINT,         OnPaintDasm);
	HANDLE_MSG(hWnd, WM_KEYDOWN,       OnKeyDownDasm);
	HANDLE_MSG(hWnd, WM_VSCROLL,       OnVScrollDasm);
	HANDLE_MSG(hWnd, WM_HSCROLL,       OnHScrollDasm);
	HANDLE_MSG(hWnd, WM_DESTROY,       OnDestroyDasm);
	HANDLE_MSG(hWnd, WM_LBUTTONUP,     OnLButtonUpDasm);
	HANDLE_MSG(hWnd, WM_RBUTTONDOWN,   OnRButtonDownDasm);
	HANDLE_MSG(hWnd, WM_LBUTTONDBLCLK, OnLButtonDblClkDasm);
	HANDLE_MSG(hWnd, WM_COMMAND,       OnCommandDasm);

	case WM_CHAR:	// forward to cmd edit window
	    windbg_cmd_fwdchar(wParam, lParam);
	    break;

	case WM_GETDLGCODE:
	    return DLGC_WANTALLKEYS /* | DLGC_WANTCHARS to get WM_CHAR messages */;

    }

    return DefWindowProc(hWnd, iMsg, wParam, lParam);
}


void
RegisterDasmClass()
{
    WNDCLASSEX  wndclass;

    wndclass.cbSize        = sizeof(wndclass);
    wndclass.style         = CS_HREDRAW | CS_VREDRAW | CS_OWNDC | CS_DBLCLKS;
    wndclass.lpfnWndProc   = WndProcDasm;
    wndclass.cbClsExtra    = 0;
    wndclass.cbWndExtra    = 0;
    wndclass.hInstance     = winstate.hInst;
    wndclass.hIcon         = LoadIcon(NULL, IDI_APPLICATION);
    wndclass.hCursor       = LoadCursor(NULL, IDC_ARROW);
    wndclass.hbrBackground = NULL; // (HBRUSH)(COLOR_WINDOW + 1);
    wndclass.lpszMenuName  = NULL;
    wndclass.lpszClassName = "solacedasm";
    wndclass.hIconSm       = LoadIcon(NULL, IDI_APPLICATION);

    RegisterClassEx(&wndclass);
}


HWND
CreateDasmWindow(HWND hwnd)
{
    DWORD winstyle = WS_CHILD | WS_VSCROLL | WS_HSCROLL | WS_VISIBLE;
    DWORD extstyle = WS_EX_CLIENTEDGE;
    HWND  hwnddasm;

    hwnddasm = CreateWindowEx(
		extstyle,		// extended style
		"solacedasm",		// window class name
		0,			// window caption
		winstyle,		// window style
		0,			// initial x position
		0,			// initial y position
		0,			// initial x size
		0,			// initial y size
		hwnd,			// parent window handle
		NULL,			// window menu handle
		winstate.hInst,		// program instance handle
		NULL);			// creation parameters

    ASSERT(hwnddasm != NULL);

    return hwnddasm;
}


// guess what the instruction prior to the one at address A is.
// this is tricky in that the previous op might start at pc-1, pc-2,
// or pc-3.  We can't always tell which as the byte at pc-1 might be
// a one byte op, or it might be an immediate operand to the 2-byte op
// at pc-2, or they both might be immediate operands to an op at pc-3.
//
// the heuristic is to move the PC back many bytes and roll forward.
// if we hit the current PC, we assume it is OK, but if the known good
// PC falls in the middle of an instruction, we assume we guessed a
// bad starting point and try another.
static word
GuessPreviousOp(word addr, int n)
{
    int guess1 = addr - (n+10)*3;  // go back n+10 max-length instructions
    int guess;
    int prev[200];
    int i;

    ASSERT((n+11)*3 < sizeof(prev)/sizeof(int));

    // don't wrap around
    if (n > addr)
	return 0;

    for(guess=guess1; guess<guess1+3; guess++) {

	int a = guess;
	int ops = 0;

	for(i=0; i<(n+11)*3; i++) {	// rolled back (n+10)*3 bytes, which might be all 1 byte or all 3 byte or a mix

	    int len = OpLen((word)a);
	    prev[ops] = a;

	    if (a == addr) {
		// we've found a plausible path to the known good op
		ASSERT(ops > n);
		return prev[ops-n];
	    }

	    a += len;
	    if (len > addr)
		break;	// we blew past the target

	    ops++;
	}
    }


    // didn't find a match!  try something else.
    //look back one, two or three bytes, looking for
    // an op that has the proper length.
    for(i=1; i<4; i++)
	if (OpLen((word)(addr-i)) == i)
	    return (word)(addr-i);


    // ok, totally confused, just decrement one byte
    return (word)(addr-1);
}


void
DasmWindowAction(win_action_t action, word addr)
{
    word dasmpc, oldpc;
    int linediff;
    int i;

    switch (action) {

    case WA_INIT:
	dasmstate.ops = NULL;
	dasmstate.active = FALSE;
	dasmstate.xoff = 0;
	dasmstate.toppc = 0;
	return;

    case WA_DISABLE:
    case WA_ENABLE:
	dasmstate.active = (action == WA_ENABLE);
	// hide/show the scroll bar
	ShowScrollBar(dbgstate.hDasmWnd, SB_VERT, dasmstate.active);
	ShowScrollBar(dbgstate.hDasmWnd, SB_HORZ, dasmstate.active);
	// force redraw
	InvalidateRect(dbgstate.hDasmWnd, NULL, TRUE);
	return;

    case WA_LINEUP:
	// make current 1st line the 2nd line.
	oldpc = dasmstate.ops[0];
	dasmpc = GuessPreviousOp((word)dasmstate.ops[0], 1);
	linediff = (dasmpc == oldpc) ? 0 : +1;
	break;

    case WA_PAGEUP:
	// make current 1st line the last line.
	dasmpc = dasmstate.ops[1];
	dasmpc = GuessPreviousOp((word)dasmstate.ops[0], dasmstate.chars_y-2);
	linediff = +999;
	break;

    case WA_LINEDOWN:
	// make current 2nd line the head of list
	dasmpc = dasmstate.ops[1];
	linediff = -1;
	break;

    case WA_PAGEDOWN:
	// make current end of list the head of list
	dasmpc = dasmstate.ops[dasmstate.chars_y-2];
	linediff = -999;
	break;

    case WA_GOTO:
	for(i=0; i<dasmstate.chars_y-1; i++) {
	    if (dasmstate.ops[i] == addr) {
		// the indicated line is already in the window, but
		// may need to redraw to indicate current position
		InvalidateRect(dbgstate.hDasmWnd, NULL, TRUE);
		return;
	    }
	}

	// start screen three instructions back to give more
	// context for current instruction.
	{
	    int opsback = min(3, dasmstate.chars_y-3);
	    dasmpc = (opsback > 0) ? GuessPreviousOp(addr, opsback)
				   : addr;
	}
	linediff = +999;	// force redraw
	break;

    default:
	return;
    }


    // save which op is on each line of the window
    // given the op of the first line of the window
    RegenerateOps(dasmpc);

    // move vertical scroll bar to indicated position
    dasmstate.toppc = dasmstate.ops[0];	// save in case of resize
    SetScrollPos(dbgstate.hDasmWnd, SB_VERT, (int)dasmstate.toppc, TRUE);

    if ((linediff > 100) || (linediff < -100))
	InvalidateRect(dbgstate.hDasmWnd, NULL, TRUE);
    else
	ScrollWindow(dbgstate.hDasmWnd,
			0, (linediff * dbgstate.fix_y),	// X,Y scroll
			NULL, NULL);
}

